/*
Header file for the Seamstress library.
*/

/*
    Copyright (C) 2007, 2008, 2009 Hayden Walles

    This file is part of Seamstress.

    Seamstress is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Seamstress is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>


/*First the declarations needed for the MATRIX ADT.*/
#ifdef USE_COMPLEX
#include <complex.h>
typedef _Complex double COMPLEX;
#endif

typedef enum {MT_BYTE,MT_INTEGER,MT_REAL,MT_COMPLEX} MATRIXTYPE;

typedef struct {
  int rows,cols;  /*Width and height*/
  MATRIXTYPE type;
  union {
    int *integers;
    double *reals;
    unsigned char *bytes;
#ifdef USE_COMPLEX
    COMPLEX *complexes;
#endif
  } buffer;
} MATRIX;


/*Error codes.*/
#define MATRIX_E_OK 0
#define MATRIX_E_MEMORY 1
#define MATRIX_E_WRONGTYPE 4

/*Macros*/
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif


#define matrixRows(m)  ((m)->rows)
#define matrixCols(m)  ((m)->cols)


#define setByte(m,r,c,i) ((m)->buffer.bytes[(r)*(m)->cols+(c)]=(i))
#define setInteger(m,r,c,i) ((m)->buffer.integers[(r)*(m)->cols+(c)]=(i))
#define setReal(m,r,c,i) ((m)->buffer.reals[(r)*(m)->cols+(c)]=(i))

#define getByte(m,r,c)((m)->buffer.bytes[(r)*(m)->cols+(c)])
#define getInteger(m,r,c) ((m)->buffer.integers[(r)*(m)->cols+(c)])
#define getReal(m,r,c) ((m)->buffer.reals[(r)*(m)->cols+(c)])

#ifdef USE_COMPLEX
#define getComplex(m,r,c) ((m)->buffer.complexes[(r)*(m)->cols+(c)])
#define setComplex(m,r,c,i) ((m)->buffer.complexes[(r)*(m)->cols+(c)]=(i))
#endif


/*Prototypes*/
void deleteMatrix(MATRIX *matrix);
MATRIX *newMatrix(int rows,int cols, int format,int *error);

/*Now the Seamstress itself.*/

/*This is the smaller NODE used by the SEAM_UNPICKER*/
/*Current maximum unpicker dimesnsions are 4096x4096*/
typedef struct {
  unsigned ergs : 20; 
  unsigned offset:12;
} NODE;
#define SEAMSTRESS_UNPICKER_MAX_SIZE (1<<12) 

/*And the larger one with both x and y offsets used for the SEAM_UNWEAVER.*/
typedef struct {
  unsigned ergs;
  unsigned short xoffset,yoffset;
} BIGNODE;


typedef MATRIX SEAM_MARKS;


typedef struct {
  int width,height;  //dimensions used for carving (same as actual values, unless direction==1 when they are swapped)
  //  IMAGE *image;   //the actual image pixel values (always actualwidth x actualheight)
  MATRIX *luminance;  //luminance information (transposed if direction==1, may also have a border)
  MATRIX *energy; //the energy values (transposed if direction==1, no border)
  SEAM_MARKS *marks;  //user annotations to the energy function (always actualwidth x actual height, if present).  This is provided by the user and should never be freed by the seams code.
  void *ergprivate;
} SEAM_COMMON;

/*A SEAM_UNPICKER holds information for a seam carving process that operates in one direction only (ie only horizontal seams or only vertical seams).  It uses some optimisations that aren't available in the more general case.*/
typedef struct {
  int actualwidth,actualheight; //dimensions of the image
  SEAM_COMMON common;
  NODE *nodes;   //the destructively updated array used to compute minimal seams (transposed if direction==1)
  int *lastseam;
  int seamcount; //number of seams removed from nodes so far
  int dynamicenergy; //flag that determines if energy is recomputed betweem seam removals
  int direction;  //flag that indicates orientation (0=vertical seams, 1=horizontal seams)
 int lastfrom,lastto; //the range of positions on the first row whose energy was altered in the last seam removal.
} SEAM_UNPICKER;

/*A SEAM_UNWEAVER holds information about a seam carving process where seams can be removed in both directions (horizontal seams and vertical seams) but isn't as efficient as the SEAM_UNPICKER.*/
typedef struct {
  SEAM_COMMON common;
  BIGNODE *nodes;   //the destructively updated array used to compute minimal seams (transposed if direction==1)
  int vseamcount,hseamcount; //number of seams removed from nodes so far
  int dynamicenergy; //flag that determines if energy is recomputed betweem seam removals
} SEAM_UNWEAVER;


typedef struct {
  int actualwidth,actualheight;  //the dimensions of the image mapped.
  int direction;  //direction (0==vertical seams, 1==horizontal seams)
  //  IMAGE *image;  //pointer to the image that is mapped.
  MATRIX *map;  //seam numbers, mapped identically to image unless direction==1 when the transpose maps identically to image
} SEAM_MAP;

/*Error codes*/
#define SEAMSTRESS_E_OK 0
#define SEAMSTRESS_E_MEMORY 1
#define SEAMSTRESS_E_BADMARKS 2
#define SEAMSTRESS_E_UNSUPPORTED 3
#define SEAMSTRESS_E_WRONGTYPE 4
#define SEAMSTRESS_E_NOENERGY 5
#define SEAMSTRESS_E_FINISHED 6
#define SEAMSTRESS_E_ABORTED 7
#define SEAMSTRESS_E_TOOBIG 8

#define SET_ERROR(a) if(error)*error=a

#define setMark(marks,row,col,type) setByte(marks,row,col,type)
#define getMark(marks,row,col) getByte(marks,row,col,type)

#define getMap(map,row,col) getInteger((map)->map,row,col)

/*convolve.c*/
MATRIX *convolvePicture(SEAM_COMMON *pic,const MATRIX *base, const MATRIX *filter, int *error);
MATRIX *dualConvolvePicture(SEAM_COMMON *pic,const MATRIX *base, const MATRIX *filter1,int norm1,const MATRIX *filter2,int norm2, int *error);
void dualConvolvePictureInc(SEAM_UNPICKER *pic,const MATRIX *base, const MATRIX *filter1,int norm1,const MATRIX *filter2,int norm2,int *froms, int *tos,MATRIX *result);

/*filters.c*/
MATRIX *createXSobel(int *error);
MATRIX *createYSobel(int *error);
MATRIX *createXGaussianGradient(double T,int *norm,int *error);
MATRIX *transpose(MATRIX *in,int *error);

/*unpicker.c*/
SEAM_UNPICKER *seamstressNewUnpicker(int actualwidth, int actualheight,int direction,int energyfunc,int dynamicenergy,int *error);
void mapDelete(SEAM_MAP *map);
SEAM_MAP *unpickerMap(SEAM_UNPICKER *pic,void (*updater)(int done,int of,void *user),void *user, int *error);
int unpickerNextSeam(SEAM_UNPICKER *pic, int *seam,int translate,int *error);
void unpickerSetRowRGB24(SEAM_UNPICKER *pic, int row, unsigned char *pixels);
void unpickerSetRowRGB32(SEAM_UNPICKER *pic, int row, unsigned char *pixels);
void unpickerDelete(SEAM_UNPICKER *pic);
void unpickerCurrentSize(SEAM_UNPICKER *picker, int *widthptr,int *heightptr);
int unpickerTransformToOriginal(SEAM_UNPICKER *pic, int *xptr, int *yptr);

/*unweaver.c*/
SEAM_UNWEAVER *seamstressNewUnweaver(int actualwidth, int actualheight,int energyfunc,int dynamicenergy,int *error);
void unweaverSetRowRGB24(SEAM_UNWEAVER *pic, int row, unsigned char *pixels);
void unweaverSetRowRGB32(SEAM_UNWEAVER *pic, int row, unsigned char *pixels);
int unweaverDelete(SEAM_UNWEAVER *weaver);
int unweaverNextVertical(SEAM_UNWEAVER *pic, int *seam,int *error);
int unweaverNextHorizontal(SEAM_UNWEAVER *pic, int *seam,int *error);

void unweaverCurrentSize(SEAM_UNWEAVER *weaver, int *widthptr,int *heightptr);
int unweaverTransformToOriginal(SEAM_UNWEAVER *weaver, int *xptr, int *yptr);

/*energy.c*/
int adjustEnergyVertically(SEAM_UNPICKER *pic,int hremove,int vremove,int *seam);
int unpickerComputeEnergy(SEAM_UNPICKER *pic,SEAM_MARKS *marks,void (*updater)(int done,int of,void *user),void *user,int *error);
int unweaverComputeEnergy(SEAM_UNWEAVER *pic,SEAM_MARKS *marks,void (*updater)(int done,int of,void *user),void *user,int *error);
/*gge.c*/
int recomputeGaussianGradientEnergy(SEAM_UNPICKER *pic,int *froms,int *tos);
int computeGaussianGradientEnergy(SEAM_COMMON *pic,void (*updater)(int done,int of,void *user),void *user,int *error);
void gaussianGradientEnergyFinish(SEAM_COMMON *pic);
int gaussianGradientEnergyStart(SEAM_COMMON *pic, int *error);
int gaussianGradientEnergyInitialise(int *error);

/*picture.c*/
void setPictureRowRGB24(MATRIX *lum,int hborder,int vborder,int transposed, int row, unsigned char *pixels);
void setPictureRowRGB32(MATRIX *lum,int hborder,int vborder,int transposed, int row, unsigned char *pixels);

/*marks.c*/
void marksDelete(SEAM_MARKS *marks);
int marksAnnotate(SEAM_MARKS *marks, int l, int t, int r, int b,int type);
SEAM_MARKS *seamstressNewMarks(int width, int height,int *error);
